# Copyright 2018 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Functions and classes involved in training'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math

from csv import DictReader, writer

import torch
import param

from pydrobert.torch.util import optimizer_to

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018 Sean Robertson"
__all__ = [
    'TrainingStateParams',
    'TrainingStateController'
]


class TrainingStateParams(param.Parameterized):
    '''Parameters controlling a TrainingStateController'''
    num_epochs = param.Integer(
        None, bounds=(1, None),
        doc='Total number of epochs to run for. If unspecified, runs '
        'until the early stopping criterion (or infinitely if disabled) '
    )
    log10_learning_rate = param.Number(
        None, softbounds=(-10, -2),
        doc='Optimizer log-learning rate. If unspecified, uses the '
        'built-in rate'
    )
    early_stopping_threshold = param.Number(
        0., bounds=(0, None), softbounds=(0, 1.),
        doc='Minimum improvement in xent from the last best that resets the '
        'early stopping clock. If zero, early stopping will not be performed'
    )
    early_stopping_patience = param.Integer(
        1, bounds=(1, None), softbounds=(1, 30),
        doc='Number of epochs where, if the classifier has failed to '
        'improve it\'s error, training is halted'
    )
    early_stopping_burnin = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs before the early stopping criterion kicks in'
    )
    reduce_lr_threshold = param.Number(
        0., bounds=(0, None), softbounds=(0, 1.),
        doc='Minimum improvement in xent from the last best that resets the '
        'clock for reducing the learning rate. If zero, the learning rate '
        'will not be reduced during training. Se'
    )
    reduce_lr_factor = param.Number(
        None, bounds=(0, 1), softbounds=(0, .5),
        inclusive_bounds=(False, False),
        doc='Factor by which to multiply the learning rate if there has '
        'been no improvement in the error after "reduce_lr_patience" '
        'epochs. If unset, uses the pytorch defaults'
    )
    reduce_lr_patience = param.Integer(
        1, bounds=(1, None), softbounds=(1, 30),
        doc='Number of epochs where, if the classifier has failed to '
        'improve it\'s error, the learning rate is reduced'
    )
    reduce_lr_cooldown = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs after reducing the learning rate before we '
        'resume checking improvements'
    )
    reduce_lr_log10_epsilon = param.Integer(
        -8, bounds=(None, 0),
        doc='The log10 absolute difference between error rates that, below '
        'which, reducing the error rate is considered meaningless'
    )
    reduce_lr_burnin = param.Integer(
        0, bounds=(0, None), softbounds=(0, 10),
        doc='Number of epochs before the criterion for reducing the learning '
        'rate kicks in'
    )
    seed = param.Integer(
        None,
        doc='Seed used for training procedures (e.g. dropout). If '
        'unset, will not touch torch\'s seeding'
    )
    dropout_prob = param.Magnitude(
        0.,
        doc='The probability of dropping a hidden unit during training'
    )
    keep_last_and_best_only = param.Boolean(
        True,
        doc='If the model is being saved, keep only the model and optimizer '
        'parameters for the last and best epoch (in terms of validation loss).'
        ' If False, save every epoch. See also "saved_model_fmt" and '
        '"saved_optimizer_fmt"'
    )
    saved_model_fmt = param.String(
        'model_{epoch:03d}.pt',
        doc='The file name format string used to save model state information.'
        ' Entries from the state csv are used to format this string (see '
        'TrainingStateController)'
    )
    saved_optimizer_fmt = param.String(
        'optim_{epoch:03d}.pt',
        doc='The file name format string used to save optimizer state '
        'information. Entries from the state csv are used to format this '
        'string (see TrainingStateController)'
    )


class TrainingStateController(object):
    '''Controls the state of training a model

    This class is used to help both control and persist experiment information
    like the current epoch, the model parameters, and model error. It assumes
    that the values stored in `params` have not changed when resuming a run.
    It is also used to control learning rates and early stopping.

    Examples
    --------

    >>> params = TrainingStateParams()
    >>> model = torch.nn.Linear(10, 1)
    >>> optimizer = torch.optim.Adam(model.parameters())
    >>> controller = TrainingStateController(
    ...    params,
    ...    state_csv_path='log.csv',
    ...    state_dir='states')
    >>> controller.load_model_and_optimizer_for_epoch(
    >>>     model, optimizer, controller.get_last_epoch())  # load previous
    >>> for epoch in range(params.num_epochs):
    ...     # ... do training loop for epoch
    ...     train_loss, val_loss = 0.1, 0.01
    ...     if not controller.update_for_epoch(
    ...             model, optimizer, train_loss, val_loss):
    ...         break  # early stopping

    Parameters
    ----------
    params : TrainingStateParams
    state_csv_path : str, optional
        A path to where training state information is stored. It stores in
        comma-separated-values format the following information. Note that
        stored values represent the state *after* updates due to epoch
        results, such as the learning rate. That way, an experiment can be
        resumed without worrying about updating the loaded results
        1. "epoch": the epoch associated with this row of information
        2. "es_resume_cd": the number of epochs left before the early
           stopping criterion begins/resumes
        3. "es_patience_cd": the number of epochs left that must pass
           without much improvement before training halts due to early stopping
        3. "rlr_resume_cd": the number of epochs left before the
           criterion for reducing the learning rate begins/resumes
        4. "rlr_patience_cd": the number of epochs left that must pass
           without much improvement before the learning rate is reduced
        5. "lr": the learning rate of the optimizer after any updates
        6. "train_met": mean training metric in exponent format. The metric
           is assumed to be lower is better
        7. "val_met": mean validation metric in exponent format. The metric
           is assumed to be lower is better
        If unset, the history will not be stored/loaded
    state_dir : str, optional
        A path to a directory to store/load model and optimizer states. If
        unset, the information will not be stored/loaded

    Attributes
    ----------
    params : TrainingParams
    state_csv_path : str or None
    state_dir : str or None
    cache_hist : dict
        A dictionary of cached results per epoch. Is not guaranteed to be
        up-to-date with `state_csv_path` unless ``update_cache()`` is called
    '''

    def __init__(self, params, state_csv_path=None, state_dir=None):
        super(TrainingStateController, self).__init__()
        self.params = params
        self.state_csv_path = state_csv_path
        self.state_dir = state_dir
        self.cache_hist = dict()

    def update_cache(self):
        '''Update the cache with history stored in state_csv_path'''
        if 0 not in self.cache_hist:
            # add a dummy entry for epoch "0" just to make logic easier. We
            # won't save it
            self.cache_hist[0] = {
                'epoch': 0,
                'es_resume_cd': self.params.early_stopping_burnin,
                'es_patience_cd': self.params.early_stopping_patience,
                'rlr_resume_cd': self.params.reduce_lr_burnin,
                'rlr_patience_cd': self.params.reduce_lr_patience,
                'train_met': float('inf'),
                'val_met': float('inf'),
                'lr': None,
            }
            if self.params.log10_learning_rate is not None:
                self.cache_hist[0]['lr'] = (
                    10 ** self.params.log10_learning_rate)
        if (self.state_csv_path is None or
                not os.path.exists(self.state_csv_path)):
            return
        with open(self.state_csv_path) as f:
            reader = DictReader(f)
            for row in reader:
                self.cache_hist[int(row['epoch'])] = {
                    'epoch': int(row['epoch']),
                    'es_resume_cd': int(row['es_resume_cd']),
                    'es_patience_cd': int(row['es_patience_cd']),
                    'rlr_resume_cd': int(row['rlr_resume_cd']),
                    'rlr_patience_cd': int(row['rlr_patience_cd']),
                    'lr': float(row['lr']),
                    'train_met': float(row['train_met']),
                    'val_met': float(row['val_met']),
                }

    def get_last_epoch(self):
        '''int : last finished epoch from training, or 0 if no history'''
        self.update_cache()
        return max(self.cache_hist)

    def get_best_epoch(self, train_met=False):
        '''Get the epoch that has lead to the best validation metric val so far

        The "best" is the lowest recorded.

        Parameters
        ----------
        train_met : bool, optional
            If ``True`` look for the best training metric value instead

        Returns
        -------
        epoch : int
            The corresponding 'best' epoch, or 0 if no epochs have run
        '''
        ent = 'train_met' if train_met else 'val_met'
        self.update_cache()
        min_epoch = 0
        min_met = self.cache_hist[0][ent]
        for info in self.cache_hist.values():
            if min_met > info[ent]:
                min_epoch = info['epoch']
                min_met = info[ent]
        return min_epoch

    def load_model_and_optimizer_for_epoch(self, model, optimizer, epoch=0):
        '''Load up model and optimizer states, or initialize them

        If `epoch` is not specified or 0, the model and optimizer are
        initialized with states for the beginning of the experiment. Otherwise,
        we look for appropriately named files in ``self.state_dir``
        '''
        model_device = next(model.parameters()).device
        if not epoch:
            # reset on cpu. Different devices can randomize differently
            model.cpu().reset_parameters()
            optim_defaults = dict(optimizer.defaults)
            if self.params.log10_learning_rate is not None:
                optim_defaults['lr'] = 10 ** self.params.log10_learning_rate
            else:
                del optim_defaults['lr']
            if self.params.seed is not None:
                torch.manual_seed(self.params.seed)
            new_optimizer = type(optimizer)(
                model.parameters(),
                **optim_defaults
            )
            model.to(model_device)
            optimizer_to(optimizer, model_device)
            optimizer.load_state_dict(new_optimizer.state_dict())
        elif self.state_dir is not None:
            epoch_info = self[epoch]
            model_basename = self.params.saved_model_fmt.format(**epoch_info)
            optimizer_basename = self.params.saved_optimizer_fmt.format(
                **epoch_info)
            model_state_dict = torch.load(
                os.path.join(self.state_dir, model_basename),
                map_location=model_device
            )
            model.load_state_dict(model_state_dict)
            optimizer_state_dict = torch.load(
                os.path.join(self.state_dir, optimizer_basename),
                map_location=model_device
            )
            optimizer.load_state_dict(optimizer_state_dict)
        else:
            print(
                'Unable to load optimizer for epoch {}. No state dict!'
                ''.format(epoch))

    def delete_model_and_optimizer_for_epoch(self, epoch):
        '''Delete state dicts for model and epoch off of disk, if they exist

        This method does nothing if the epoch records or the files do not
        exist. It is called during ``update_for_epoch`` if the parameter
        ``keep_last_and_best_only`` is ``True``

        Parameters
        ----------
        epoch : int
            The epoch in question
        '''
        if self.state_dir is None:
            return
        epoch_info = self.get_info(epoch, None)
        if epoch_info is None:
            return
        model_basename = self.params.saved_model_fmt.format(**epoch_info)
        optimizer_basename = self.params.saved_optimizer_fmt.format(
            **epoch_info)
        try:
            os.remove(os.path.join(self.state_dir, model_basename))
        except OSError as e:
            pass
        try:
            os.remove(os.path.join(self.state_dir, optimizer_basename))
        except OSError:
            pass

    def get_info(self, epoch, *default):
        '''Get history entries for a specific epoch

        If there's an entry present for `epoch`, return it. The value is a
        dictionary with the keys "epoch", "es_resume_cd", "es_patience_cd",
        "rlr_resume_cd", "rlr_patience_cd", "lr", "train_met", and "val_met".

        If there's no entry for `epoch`, and no additional arguments were
        passed to this method, it raises a ``KeyError``. If an additional
        argument was passed to this method, return it.
        '''
        if len(default) > 1:
            raise TypeError('expected at most 2 arguments, got 3')
        if epoch in self.cache_hist:
            return self.cache_hist[epoch]
        self.update_cache()
        return self.cache_hist.get(epoch, *default)

    def __getitem__(self, epoch):
        return self.get_info(epoch)

    def save_model_and_optimizer_with_info(self, model, optimizer, info):
        '''Save model and optimizer state dictionaries to file given epoch info

        This is called automatically during ``update_for_epoch``. Does not save
        if there is no directory to save to (i.e. ``self.state_dir is None``).
        Format strings from ``self.params`` are formatted with the values from
        `info` to construct the base names of each file

        Parameters
        ----------
        model : AcousticModel
        optimizer : torch.optim.Optimizer
        info : dict
            A dictionary with the entries "epoch", "es_resume_cd",
            "es_patience_cd", "rlr_resume_cd", "rlr_patience_cd", "lr",
            "train_met", and "val_met"
        '''
        if self.state_dir is None:
            return
        if not os.path.isdir(self.state_dir):
            os.makedirs(self.state_dir)
        model_basename = self.params.saved_model_fmt.format(**info)
        optimizer_basename = self.params.saved_optimizer_fmt.format(
            **info)
        model_state_dict = model.state_dict()
        # we always save on the cpu
        for key, val in model_state_dict.items():
            model_state_dict[key] = val.cpu()
        optimizer_to(optimizer, 'cpu')
        torch.save(
            model.state_dict(),
            os.path.join(self.state_dir, model_basename),
        )
        torch.save(
            optimizer.state_dict(),
            os.path.join(self.state_dir, optimizer_basename),
        )

    def save_info_to_hist(self, info):
        '''Append history entries to the history csv

        This is called automatically during ``update_for_epoch``. Does not save
        if there is no file to save to (i.e. ``self.state_csv_path is None``).
        Values are appended to the end of the csv file - no checking is
        performed for mantaining a valid history.

        Parameters
        ----------
        info : dict
            A dictionary with the entries "epoch", "es_resume_cd",
            "es_patience_cd", "rlr_resume_cd", "rlr_patience_cd", "lr",
            "train_met", and "val_met"
        '''
        self.cache_hist[info['epoch']] = info
        if self.state_csv_path is None:
            return
        if not self.params.num_epochs:
            epoch_fmt_str = '{:010d}'
        else:
            epoch_fmt_str = '{{:0{}d}}'.format(
                int(math.log10(self.params.num_epochs)) + 1)
        es_resume_cd_fmt_str = '{{:0{}d}}'.format(
            int(math.log10(max(
                self.params.early_stopping_burnin,
                1,
                ))) + 1
        )
        es_patience_cd_fmt_str = '{{:0{}d}}'.format(
            int(math.log10(max(
                self.params.early_stopping_patience,
                1,
                ))) + 1
        )
        rlr_resume_cd_fmt_str = '{{:0{}d}}'.format(
            int(math.log10(max(
                self.params.reduce_lr_cooldown,
                self.params.reduce_lr_burnin,
                1,
            ))) + 1
        )
        rlr_patience_cd_fmt_str = '{{:0{}d}}'.format(
            int(math.log10(max(
                self.params.reduce_lr_patience,
                1,
            ))) + 1
        )
        lr_fmt_str = train_met_fmt_str = val_met_fmt_str = '{:10e}'
        write_header = not os.path.exists(self.state_csv_path)
        with open(self.state_csv_path, 'a') as f:
            wr = writer(f)
            if write_header:
                wr.writerow([
                    'epoch',
                    'es_resume_cd',
                    'es_patience_cd',
                    'rlr_resume_cd',
                    'rlr_patience_cd',
                    'lr',
                    'train_met',
                    'val_met',
                ])
            wr.writerow([
                epoch_fmt_str.format(info['epoch']),
                es_resume_cd_fmt_str.format(info['es_resume_cd']),
                es_patience_cd_fmt_str.format(info['es_patience_cd']),
                rlr_resume_cd_fmt_str.format(info['rlr_resume_cd']),
                rlr_patience_cd_fmt_str.format(info['rlr_patience_cd']),
                lr_fmt_str.format(info['lr']),
                train_met_fmt_str.format(info['train_met']),
                val_met_fmt_str.format(info['val_met']),
            ])

    def update_for_epoch(
            self, model, optimizer, train_met, val_met, epoch=None):
        '''Update history and optimizer after latest epoch results

        Parameters
        ----------
        model : AcousticModel
        optimizer : torch.optim.Optimizer
        train_met : float
            Mean value of metric on training set for epoch
        val_met : float
            Mean value of metric on validation set for epoch
        epoch : int, optional
            The epoch that just finished. If unset, it is inferred to be one
            after the last epoch in the history

        Returns
        -------
        continue_training : bool
            Whether to continue training. This can be set to ``False`` either
            by hitting the max number of epochs or by early stopping
        '''
        if epoch is None:
            epoch = self.get_last_epoch() + 1
        if not self.params.num_epochs:
            continue_training = True
        else:
            continue_training = epoch < self.params.num_epochs
        info = dict(self.get_info(epoch - 1, None))
        if info is None:
            raise ValueError(
                'no entry for the previous epoch {}, so unable to update'
                ''.format(epoch))
        last_best = self.get_best_epoch()
        best_info = self[last_best]
        if info['lr'] is None:
            # can only happen during the first epoch. We don't know the
            # optimizer defaults, so we get them now
            info['lr'] = optimizer.defaults['lr']
        if info["es_resume_cd"]:
            info["es_resume_cd"] -= 1
        elif (max(best_info['val_met'] - val_met, 0) <
                self.params.early_stopping_threshold):
            info["es_patience_cd"] -= 1
            if not info["es_patience_cd"]:
                continue_training = False
        else:
            info["es_patience_cd"] = self.params.early_stopping_patience
        if info["rlr_resume_cd"]:
            info["rlr_resume_cd"] -= 1
        elif (max(best_info['val_met'] - val_met, 0) <
                self.params.reduce_lr_threshold):
            info["rlr_patience_cd"] -= 1
            if not info["rlr_patience_cd"]:
                old_lr = info['lr']
                new_lr = old_lr * self.params.reduce_lr_factor
                rlr_epsilon = 10 ** self.params.reduce_lr_log10_epsilon
                if old_lr - new_lr > rlr_epsilon:
                    info['lr'] = new_lr
                    for param_group in optimizer.param_groups:
                        # just assume that the user knows what's what if
                        # the optimizer's lr doesn't match the old one
                        param_group['lr'] = new_lr
                info["rlr_resume_cd"] = self.params.reduce_lr_cooldown
                info["rlr_patience_cd"] = self.params.reduce_lr_patience
        else:
            info["rlr_patience_cd"] = self.params.reduce_lr_patience
        info["epoch"] = epoch
        info["val_met"] = val_met
        info["train_met"] = train_met
        self.save_model_and_optimizer_with_info(model, optimizer, info)
        self.save_info_to_hist(info)
        cur_best = self.get_best_epoch()
        if self.params.keep_last_and_best_only and cur_best != epoch - 1:
            self.delete_model_and_optimizer_for_epoch(epoch - 1)
        if self.params.keep_last_and_best_only and cur_best != last_best:
            self.delete_model_and_optimizer_for_epoch(last_best)
        return continue_training



# Copyright 2018 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Utility functions'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018 Sean Robertson"
__all__ = [
    'optimizer_to',
]


def optimizer_to(optimizer, to):
    '''Move tensors in an optimizer to another device

    This function traverses the state dictionary of an `optimizer` and pushes
    any tensors it finds to the device implied by `to` (as per the semantics
    of ``torch.tensor.to``)
    '''
    state_dict = optimizer.state_dict()
    key_stack = [(x,) for x in state_dict.keys()]
    new_device = False
    while key_stack:
        last_dict = None
        val = state_dict
        keys = key_stack.pop()
        for key in keys:
            last_dict = val
            val = val[key]
        if isinstance(val, torch.Tensor):
            try:
                last_dict[keys[-1]] = val.to(to)
                if last_dict[keys[-1]].device != val.device:
                    new_device = True
            except Exception as e:
                print(e)
                pass
        elif hasattr(val, 'keys'):
            key_stack += [keys + (x,) for x in val.keys()]
    if new_device:
        optimizer.load_state_dict(state_dict)



# Copyright 2018 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import argparse
import math

import torch
import pydrobert.torch.data as data

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018 Sean Robertson"
__all__ = [
    'get_torch_spect_data_dir_info',
]


def _get_torch_spect_data_dir_info_parse_args(args):
    parser = argparse.ArgumentParser(
        description=get_torch_spect_data_dir_info.__doc__,
    )
    parser.add_argument('dir', type=str, help='The torch data directory')
    parser.add_argument(
        'out_file', nargs='?', type=argparse.FileType('w'),
        default=sys.stdout,
        help='The file to write to. If unspecified, stdout',
    )
    parser.add_argument(
        '--strict', action='store_true', default=False,
        help='If set, validate the data directory before collecting info. The '
        'process is described in pydrobert.torch.data.validate_spect_data_set'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    return parser.parse_args(args)


def get_torch_spect_data_dir_info(args=None):
    '''Write info about the specified SpectDataSet data dir

    A torch ``SpectDataSet`` data dir is of the form::

        dir/
            feats/
                <file_prefix><utt1><file_suffix>
                <file_prefix><utt2><file_suffix>
                ...
            [ali/
                <file_prefix><utt1><file_suffix>
                <file_prefix><utt1><file_suffix>
                ...
            ]

    Where ``feats`` contains ``FloatTensor``s of shape ``(N, F)``, where
    ``N`` is the number of frames (variable) and ``F`` is the number of
    filters (fixed) and ``ali``, if there, contains ``LongTensor``s of shape
    ``(N,)`` indicating the appropriate class labels.

    This command writes the following space-delimited key-value pairs to an
    output file in sorted order:

    1. "num_utterances", the total number of listed utterances
    2. "num_filts", ``F``
    3. "total_frames", ``sum(N)`` over the data dir
    4. "count_<i>", the number of instances of the class "<i>" that appear
        in ``ali`` (if available). If "count_<i>" is a valid key, then so
        are "count_<0 to i>". "count_<i>" is left-padded with zeros to ensure
        that the keys remain in the same order in the table as the class
        indices

    Note that the output can be parsed as a Kaldi text table of integers.
    '''
    try:
        options = _get_torch_spect_data_dir_info_parse_args(args)
    except SystemExit as ex:
        return ex.code
    if not os.path.isdir(options.dir):
        print("'{}' is not a directory".format(options.dir), file=sys.stderr)
        return 1
    data_set = data.SpectDataSet(
        options.dir,
        file_prefix=options.file_prefix,
        file_suffix=options.file_suffix,
    )
    if options.strict:
        data.validate_spect_data_set(data_set)
    info_dict = {
        'num_utterances': len(data_set),
        'total_frames': 0
    }
    counts = dict()
    max_class_idx = -1
    for feat, ali in data_set:
        info_dict['num_filts'] = feat.size()[1]
        info_dict['total_frames'] += feat.size()[0]
        if ali is not None:
            for class_idx in ali:
                class_idx = class_idx.item()
                max_class_idx = max(class_idx, max_class_idx)
                if max_class_idx < 0:
                    raise ValueError('Got a negative class idx')
                counts[class_idx] = counts.get(class_idx, 0) + 1
    if max_class_idx == 0:
        info_dict['count_0'] = counts[0]
    elif max_class_idx > 0:
        count_fmt_str = 'count_{{:0{}d}}'.format(
            int(math.log10(max_class_idx)) + 1)
        for class_idx in range(max_class_idx + 1):
            info_dict[count_fmt_str.format(class_idx)] = counts.get(
                class_idx, 0)
    info_list = sorted(info_dict.items())
    for key, value in info_list:
        options.out_file.write("{} {}\n".format(key, value))
    if options.out_file != sys.stdout:
        options.out_file.close()
    return 0



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from codecs import open
from os import path
from setuptools import setup

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018 Sean Robertson"

PWD = path.abspath(path.dirname(__file__))
with open(path.join(PWD, 'README.md'), encoding='utf-8') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

SETUP_REQUIRES = ['setuptools_scm']
if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
    SETUP_REQUIRES += ['pytest-runner']


def main():
    setup(
        name='pydrobert-pytorch',
        description='PyTorch utilities for ML, specifically speech',
        long_description=LONG_DESCRIPTION,
        use_scm_version=True,
        zip_safe=False,
        url='https://github.com/sdrobert/pydrobert-pytorch',
        author=__author__,
        author_email=__email__,
        license=__license__,
        packages=['pydrobert', 'pydrobert.torch'],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],
        install_requires=[
            'numpy', 'future', 'torch', 'param'
        ],
        setup_requires=SETUP_REQUIRES,
        tests_require=[
            'pytest', 'scipy',
        ],
        entry_points={
            'console_scripts': [
                'get-torch-spect-data-dir-info = pydrobert.torch.command_line:'
                'get_torch_spect_data_dir_info'
            ]
        },
    )


if __name__ == '__main__':
    main()



'''Classes and functions related to storing/retrieving data'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import warnings

import numpy as np
import torch
import torch.utils.data
import param

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2018 Sean Robertson"
__all__ = [
    'extract_window',
    'SpectDataSet',
    'validate_spect_data_set',
    'UtteranceContextWindowDataSet',
    'SingleContextWindowDataSet',
    'EpochRandomSampler',
    'context_window_seq_to_batch',
    'DataSetParams',
    'SpectDataParams',
    'SpectDataSetParams',
    'TrainingDataLoader',
    'EvaluationDataLoader',
]


def extract_window(signal, frame_idx, left, right):
    '''Slice the signal to extract a context window

    Parameters
    ----------
    signal : torch.Tensor
        Of shape ``(T, F)``, where ``T`` is the time/frame axis, and ``F``
        is the frequency axis
    frame_idx : int
        The "center frame" ``0 <= frame_idx < T``
    left : int
        The number of frames in the window to the left (before) the center
        frame. Any frames below zero are edge-padded
    right : int
        The number of frames in the window to the right (after) the center
        frame. Any frames above ``T`` are edge-padded

    Returns
    -------
    window : torch.Tensor
        Of shape ``(1 + left + right, F)``
    '''
    T, F = signal.size()
    if frame_idx - left < 0 or frame_idx + right + 1 > T:
        win_size = 1 + left + right
        window = signal.new(win_size, F)
        left_pad = max(left - frame_idx, 0)
        right_pad = max(frame_idx + right + 1 - T, 0)
        window[left_pad:win_size - right_pad] = signal[
            max(0, frame_idx - left):frame_idx + right + 1]
        if left_pad:
            window[:left_pad] = signal[0]
        if right_pad:
            window[-right_pad:] = signal[-1]
    else:
        window = signal[frame_idx - left:frame_idx + right + 1]
    return window


class SpectDataSet(torch.utils.data.Dataset):
    '''Accesses spectrographic filter data stored in a data directory

    ``SpectDataSet`` assumes that `data_dir` is structured as

    ::
        data_dir/
            feats/
                <file_prefix><uttid1><file_suffix>
                <file_prefix><uttid2><file_suffix>
                ...
            [
            ali/
                <file_prefix><uttid1><file_suffix>
                <file_prefix><uttid2><file_suffix>
                ...
            ]

    The ``feats`` dir stores filter bank data in the form of
    ``torch.FloatTensor``s of size ``(T, F)``, where ``T`` is the time
    dimension and ``F`` is the filter/log-frequency dimension. ``ali`` stores
    ``torch.LongTensor`s of size ``(T,)`` indicating the pdf-id of the most
    likely target. ``ali/`` is optional.

    As a ``Sequence``, the ``i``th index of a ``SpectDataSet`` instance
    returns a pair of ``feat[uttids[i]], ali[uttids[i]]``. If ``ali/`` did not
    exist on initialization, the second element of the pair will be ``None``

    Parameters
    ----------
    data_dir : str
        A path to feature directory
    file_prefix : str, optional
        The prefix that indicates that the file counts toward the data set
    file_suffix : str, optional
        The suffix that indicates that the file counts toward the data set
    warn_on_missing : bool, optional
        If some files with ``file_suffix`` exist in the ``ali/`` dir,
        there's a mismatch between the utterances in ``feats/`` and ``ali/``,
        and `warn_on_missing` is ``True``, a warning will be issued
        (via ``warnings``) regarding each such mismatch

    Attributes
    ----------
    data_dir : str
    file_suffix : str
    has_ali : bool
        Whether alignment data exist
    utt_ids : tuple
        A tuple of all utterance ids extracted from the data directory. They
        are stored in the same order as features and alignments via
        ``__getitem__``. If the ``ali/`` directory exists, `utt_ids`
        contains only the utterances in the intersection of each directory.
    '''

    def __init__(
            self, data_dir, file_prefix='', file_suffix='.pt',
            warn_on_missing=True):
        super(SpectDataSet, self).__init__()
        self.data_dir = data_dir
        self.file_prefix = file_prefix
        self.file_suffix = file_suffix
        self.has_ali = os.path.isdir(os.path.join(data_dir, 'ali'))
        if self.has_ali:
            self.has_ali = bool(sum(
                1 for x in os.listdir(os.path.join(data_dir, 'ali'))
                if x.startswith(file_prefix) and x.endswith(file_suffix)
            ))
        self.utt_ids = tuple(sorted(
            self.find_utt_ids(warn_on_missing)))

    def __len__(self):
        return len(self.utt_ids)

    def __getitem__(self, idx):
        return self.get_utterance_pair(idx)

    def find_utt_ids(self, warn_on_missing):
        '''Iterator : all utterance ids from data_dir'''
        neg_fsl = -len(self.file_suffix)
        if not neg_fsl:
            neg_fsl = None
        fpl = len(self.file_prefix)
        utt_ids = (
            os.path.basename(x)[fpl:neg_fsl]
            for x in os.listdir(os.path.join(self.data_dir, 'feats'))
            if x.startswith(self.file_prefix) and x.endswith(self.file_suffix)
        )
        try:
            ali_utt_ids = set(
                os.path.basename(x)[fpl:neg_fsl]
                for x in os.listdir(os.path.join(self.data_dir, 'ali'))
                if x.startswith(self.file_prefix) and
                x.endswith(self.file_suffix)
            )
        except OSError:
            assert not self.has_ali
            ali_utt_ids = set()
        if ali_utt_ids:
            utt_ids = set(utt_ids)
            if warn_on_missing:
                for utt_id in utt_ids.difference(ali_utt_ids):
                    warnings.warn("Missing ali for uttid: '{}'".format(utt_id))
                for utt_id in ali_utt_ids.difference(utt_ids):
                    warnings.warn(
                        "Missing feats for uttid: '{}'".format(utt_id))
            utt_ids &= ali_utt_ids
        return utt_ids

    def get_utterance_pair(self, idx):
        '''Get a pair of features, alignments'''
        utt_id = self.utt_ids[idx]
        feats = torch.load(
            os.path.join(
                self.data_dir,
                'feats',
                self.file_prefix + utt_id + self.file_suffix))
        if self.has_ali:
            ali = torch.load(
                os.path.join(
                    self.data_dir,
                    'ali',
                    self.file_prefix + utt_id + self.file_suffix))
        else:
            ali = None
        return feats, ali

    def write_pdf(self, utt, pdf):
        '''Write a pdf FloatTensor to the data directory

        This method writes a pdf matrix to the directory ``data_dir/pdfs``
        with the name ``<file_prefix><utt><file_suffix>``

        Parameters
        ----------
        utt : str or int
            The name of the utterance to write. If an integer is specified,
            `utt` is assumed to index an utterance id specified in
            ``self.utt_ids``
        pdf : torch.Tensor
            The tensor to write. It will be converted to a ``FloatTensor``
            using the command ``pdf.cpu().float()``
        '''
        if isinstance(utt, int):
            utt = self.utt_ids[utt]
        pdfs_dir = os.path.join(self.data_dir, 'pdfs')
        if not os.path.isdir(pdfs_dir):
            os.makedirs(pdfs_dir)
        torch.save(
            pdf.cpu().float(),
            os.path.join(pdfs_dir, self.file_prefix + utt + self.file_suffix)
        )


def validate_spect_data_set(data_set):
    '''Validate SpectDataSet data directory

    The data directory is valid if the following conditions are observed

     1. All features are ``FloatTensor`` instances
     2. All features have two axes
     3. All features have the same size second axis
     4. All alignments (if present) are ``LongTensor`` instances
     5. All alignments (if present) have one axis
     6. Features and alignments (if present) have the same size first
        axes for a given utterance id

    Raises a ``ValueError`` if a condition is violated
    '''
    num_filts = None
    for idx in range(len(data_set.utt_ids)):
        feats, ali = data_set.get_utterance_pair(idx)
        if not isinstance(feats, torch.FloatTensor):
            raise ValueError(
                "'{}' (index {}) in '{}' is not a FloatTensor".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'feats')))
        if len(feats.size()) != 2:
            raise ValueError(
                "'{}' (index {}) in '{}' does not have two axes".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'feats')
                ))
        if num_filts is None:
            num_filts = feats.size()[1]
        elif feats.size()[1] != num_filts:
            raise ValueError(
                "'{}' (index {}) in '{}' has second axis size {}, which "
                "does not match prior utterance ('{}') size of {}".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'feats'),
                    feats.size()[1],
                    data_set.utt_ids[idx - 1] + data_set.file_suffix,
                    num_filts))
        if ali is None:
            continue
        if not isinstance(ali, torch.LongTensor):
            raise ValueError(
                "'{}' (index {}) in '{}' is not a LongTensor".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'ali')))
        if len(ali.size()) != 1:
            raise ValueError(
                "'{}' (index {}) in '{}' does not have one axis".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'ali')))
        if ali.size()[0] != feats.size()[0]:
            raise ValueError(
                "'{}' (index {}) in '{}' does not have the same first axis "
                "size ({}) as it's companion in '{}' ({})".format(
                    data_set.utt_ids[idx] + data_set.file_suffix, idx,
                    os.path.join(data_set.data_dir, 'feats'),
                    feats.size()[0],
                    os.path.join(data_set.data_dir, 'ali'),
                    ali.size()[0]))


class UtteranceContextWindowDataSet(SpectDataSet):
    '''SpectDataSet, extracting fixed-width windows over the utterance

    Like a ``SpectDataSet``, ``UtteranceContextWindowDataSet`` indexes pairs of
    features and alignments. Instead of returning features of shape ``(T, F)``,
    instances return features of shape ``(T, 1 + left + right, F)``, where the
    ``T`` axis indexes the so-called center frame and the ``1 + left + right``
    axis contains frame vectors (size ``F``) including the center frame,
    ``left`` frames in time before the center frame, and ``right`` frames after

    Parameters
    ----------
    data_dir : str
    left : int
    right : int
    file_prefix : str, optional
    file_suffix : str, optional
    warn_on_missing : bool, optional

    Attributes
    ----------
    data_dir : str
    left : int
    right : int
    has_ali : bool
    utt_ids : tuple
    '''

    def __init__(self, data_dir, left, right, **kwargs):
        super(UtteranceContextWindowDataSet, self).__init__(data_dir, **kwargs)
        self.left = left
        self.right = right

    def get_windowed_utterance(self, idx):
        '''Get pair of features (w/ context window) and alignments'''
        feats, ali = self.get_utterance_pair(idx)
        num_frames, num_filts = feats.size()
        windowed = torch.empty(
            num_frames, 1 + self.left + self.right, num_filts)
        for center_frame in range(num_frames):
            windowed[center_frame] = extract_window(
                feats, center_frame, self.left, self.right)
        return windowed, ali

    def __getitem__(self, idx):
        return self.get_windowed_utterance(idx)


class SingleContextWindowDataSet(SpectDataSet):
    '''SpectDataSet, returning a single context window per index

    Like ``SpectDataSet``, ``SingleContextWindowDataSet`` indexes pairs of
    features and alignments. Instead of indexing features of shape ``(T, F)``
    and alignments of shape ``T``, instances break down each utterance into
    ``T`` separate context windows and ``T`` integers and indexes those. For
    the ``t``-th context window of features, we have a tensor of size ``(1 +
    left + right, F)``, where ``window[left]`` is the "center frame",
    ``window[:left]`` are the frames before it, and ``window[left + 1:]`` are
    those after. Context windows are ordered first by utterance, second by
    center frame idx. For example, if there are only two utterances, "a" and
    "b" of sizes ``(2, 5)`` and ``(4, 5)``, respectively, then the first index
    (0) of the data set would be the context window of "a"'s first frame,
    whereas the 4th index (3) of the data set points to the context window of
    "b"'s second frame (1).

    Attributes
    ----------
    data_dir : str
    left : int
    right : int
    has_ali : bool
    utt_ids : tuple
    utt_lens : tuple
        A tuple containing the number of frames of each utterance (i.e. ``T``)
    '''

    def __init__(self, data_dir, left, right, **kwargs):
        super(SingleContextWindowDataSet, self).__init__(data_dir, **kwargs)
        self.left = left
        self.right = right
        self.utt_lens = tuple(
            torch.load(os.path.join(
                self.data_dir, 'feats', x + self.file_suffix)).size()[0]
            for x in self.utt_ids
        )

    def __len__(self):
        return sum(self.utt_lens)

    def get_context_window(self, idx):
        idx_err_msg = 'list index out of range'
        if idx < 0:
            idx += len(self)
            if idx < 0:
                raise IndexError(idx_err_msg)
        utt_idx = 0
        for utt_len in self.utt_lens:
            if idx < utt_len:
                break
            idx -= utt_len
            utt_idx += 1
        if utt_idx == len(self.utt_ids):
            raise IndexError(idx_err_msg)
        feats, ali = self.get_utterance_pair(utt_idx)
        window = extract_window(feats, idx, self.left, self.right)
        return window, ali[idx].item() if ali is not None else ali

    def __getitem__(self, idx):
        return self.get_context_window(idx)


class EpochRandomSampler(torch.utils.data.Sampler):
    '''Return random samples that are the same for a fixed epoch

    Parameters
    ----------
    data_source : torch.data.utils.Dataset
        The total number of samples
    init_epoch : int, optional
        The initial epoch
    base_seed : int, optional
        Determines the starting seed of the sampler. Sampling is seeded with
        ``base_seed + epoch``. If unset, a seed is randomly generated from
        the default generator

    Attributes
    ----------
    epoch : int
        The current epoch. Responsible for seeding the upcoming samples
    data_source : torch.data.utils.Dataset
    base_seed : int

    Examples
    --------
    >>> sampler = EpochRandomSampler(
    ...     torch.data.utils.TensorDataset(torch.arange(100)))
    >>> samples_ep0 = tuple(sampler)  # random
    >>> samples_ep1 = tuple(sampler)  # random, probably not same as first
    >>> assert tuple(sampler.get_samples_for_epoch(0)) == samples_ep0
    >>> assert tuple(sampler.get_samples_for_epoch(1)) == samples_ep1
    '''

    def __init__(self, data_source, init_epoch=0, base_seed=None):
        super(EpochRandomSampler, self).__init__(data_source)
        self.data_source = data_source
        self.epoch = init_epoch
        if base_seed is None:
            base_seed = np.random.randint(np.iinfo(np.uint32).max)
        self.base_seed = base_seed

    def __len__(self):
        return len(self.data_source)

    def get_samples_for_epoch(self, epoch):
        '''tuple : samples for a specific epoch'''
        rs = np.random.RandomState(self.base_seed + epoch)
        return rs.permutation(range(len(self.data_source)))

    def __iter__(self):
        ret = iter(self.get_samples_for_epoch(self.epoch))
        self.epoch += 1
        return ret


def context_window_seq_to_batch(seq):
    '''Convert a sequence of context window elements to a batch

    Given a sequence of ``feats, ali`` pairs, where ``feats`` is either of size
    ``(T, C, F)`` or ``(C, F)``, where ``T`` is some number of windows (which
    can vary across elements in the sequence), ``C`` is the window size, and
    ``F`` is some number filters, and ``ali`` is either of size ``(T,)`` or is
    an integer, this method batches all the elements of the sequence into a
    pair of ``batch_feats, batch_ali``, where ``batch_feats`` is of size
    ``(N, C, F)``, ``N`` being the total number of windows, and ``batch_ali``
    is of size ``(N,)``

    Parameters
    ----------
    seq : sequence

    Returns
    -------
    batch_feats, batch_ali : torch.FloatTensor, torch.LongTensor or None
    '''
    batch_feats = []
    batch_ali = []
    for feats, ali in seq:
        if len(feats.size()) == 2:
            feats = feats.unsqueeze(0)
        batch_feats.append(feats)
        if ali is None:
            # assume every remaining ali will be none
            batch_ali = None
        else:
            if isinstance(ali, int):
                ali = torch.tensor([ali])
            batch_ali.append(ali)
    batch_feats = torch.cat(batch_feats)
    if batch_ali is not None:
        batch_ali = torch.cat(batch_ali)
    return batch_feats, batch_ali


class DataSetParams(param.Parameterized):
    '''General parameters for a single partition of data'''
    batch_size = param.Integer(
        10, bounds=(1, None),
        doc='Number of elements in a batch. For training, this is the total '
        'number of context windows in a batch. For validation/testing, this '
        'is the number of utterances to process at once'
    )
    seed = param.Integer(
        None,
        doc='The seed used to shuffle data. The seed is incremented at every '
        'epoch'
    )
    drop_last = param.Boolean(
        False,
        doc='Whether to drop the last batch if it does reach batch_size'
    )


class SpectDataParams(param.Parameterized):
    '''Parameters for spectral data'''
    # context windows are more model parameters than data parameters, but
    # we're going to extract them as part of the data loading process, which
    # is easily parallelized by the DataLoader
    context_left = param.Integer(
        4, bounds=(0, None), softbounds=(3, 8),
        doc='How many frames to the left of (before) the current frame are '
        'included when determining the class of the current frame'
    )
    context_right = param.Integer(
        4, bounds=(0, None), softbounds=(3, 8),
        doc='How many frames to the right of (after) the current frame are '
        'included when determining the class of the current frame'
    )


class SpectDataSetParams(SpectDataParams, DataSetParams):
    '''Data set parameters for a specific partition of spectral data'''
    pass


class TrainingDataLoader(torch.utils.data.DataLoader):
    '''Serve batches of context windows randomly for training

    Parameters
    ----------
    data_dir : str
        Path to the torch data directory. Should have the format
        ::
            data_dir/
                feats/
                    <file_prefix><utt1><file_suffix>
                    <file_prefix><utt2><file_suffix>
                    ...
                ali/
                    <file_prefix><utt1><file_suffix>
                    <file_prefix><utt2><file_suffix>
                    ...
    params : SpectDataSetParams
        Parameters for things like context window size, batch size, and
        seed. For this loader, the batch size equals the number of context
        windows
    file_prefix : str, optional
        The prefix that indicates that the file counts toward the data set
    file_suffix : str, optional
        The suffix that indicates that the file counts toward the data set
    warn_on_missing : bool, optional
        If some files with ``file_suffix`` exist in the ``ali/`` dir,
        there's a mismatch between the utterances in ``feats/`` and ``ali/``,
        and `warn_on_missing` is ``True``, a warning will be issued
        (via ``warnings``) regarding each such mismatch
    init_epoch : int, optional
        Where training should resume from
    kwargs : keyword arguments, optional
        Additional ``DataLoader`` arguments

    Attributes
    ----------
    data_dir : str
    params : cnn_model.params.SpectDataParams

    Yields
    ------
    feats, alis : tensor, tensor, tuple, tuple
        `feats` is a ``FloatTensor`` of size ``(params.batch_size, C, F)``,
        where ``param.batch_size`` is the number of context windows, ``C`` is
        the context window size, and ``F`` is the number of filters per frame.
        `ali` is a ``LongTensor`` of size ``(params.batch_size,)``
    '''

    def __init__(
            self, data_dir, params, init_epoch=0, file_prefix='',
            file_suffix='.pt', warn_on_missing=True, **kwargs):
        for bad_kwarg in (
                'batch_size', 'sampler', 'batch_sampler', 'shuffle',
                'collate_fn'):
            if bad_kwarg in kwargs:
                raise TypeError(
                    'keyword argument "{}" invalid for {} types'.format(
                        bad_kwarg, type(self)))
        self.data_dir = data_dir
        self.params = params
        self.data_source = SingleContextWindowDataSet(
            data_dir, params.context_left, params.context_right,
            file_prefix=file_prefix, file_suffix=file_suffix,
            warn_on_missing=warn_on_missing,
        )
        if not self.data_source.has_ali:
            raise ValueError(
                "'{}' must have alignment info for training".format(
                    data_dir))
        self.__sampler = EpochRandomSampler(
            self.data_source, init_epoch=init_epoch, base_seed=params.seed)
        batch_sampler = torch.utils.data.BatchSampler(
            self.__sampler, params.batch_size, drop_last=params.drop_last)
        super(TrainingDataLoader, self).__init__(
            self.data_source,
            batch_sampler=batch_sampler,
            collate_fn=context_window_seq_to_batch,
            **kwargs
        )

    @property
    def epoch(self):
        '''int : the current epoch'''
        return self.__sampler.epoch

    @epoch.setter
    def epoch(self, val):
        self.__sampler.epoch = val


class EvaluationDataLoader(torch.utils.data.DataLoader):
    '''Serves batches of context windows over sequential utterances

    Parameters
    ----------
    data_dir : str
        Path to the torch data directory. Should have the format
        ::
            data_dir/
                feats/
                    <file_prefix><utt1><file_suffix>
                    <file_prefix><utt2><file_suffix>
                    ...
                [ali/
                    <file_prefix><utt1><file_suffix>
                    <file_prefix><utt2><file_suffix>
                    ...]
        Where the alignment directory is optional
    params : SpectDataSetParams
        Parameters for things like context window size, batch size, and
        seed. For this loader, the batch size equals the number of utterances
    file_prefix : str, optional
        The prefix that indicates that the file counts toward the data set
    file_suffix : str, optional
        The suffix that indicates that the file counts toward the data set
    warn_on_missing : bool, optional
        If some files with ``file_suffix`` exist in the ``ali/`` dir,
        there's a mismatch between the utterances in ``feats/`` and ``ali/``,
        and `warn_on_missing` is ``True``, a warning will be issued
        (via ``warnings``) regarding each such mismatch
    kwargs : keyword arguments, optional
        Additional ``DataLoader`` arguments

    Attributes
    ----------
    data_dir : str
    params : cnn_model.params.SpectDataParams

    Yields
    ------
    feats, alis, feat_sizes, utt_ids : tensor, tensor, tuple, tuple
        `feats` is a ``FloatTensor`` of size ``(N, C, F)``, where ``N`` is
        the number of context windows, ``C`` is the context window size, and
        ``F`` is the number of filters per frame. `ali` is a ``LongTensor``
        of size ``(N,)`` (or ``None`` if the ``ali`` dir was not specified).
        `feat_sizes` is a tuple of size ``params.batch_size``
        specifying the number of context windows per utterance in the batch.
        ``feats[sum(feat_sizes[:i]):sum(feat_sizes[:i+1])]`` are the context
        windows for the ``i``-th utterance in the batch
        (``sum(feat_sizes) == N``). ``utt_ids`` is a tuple of size
        ``params.batch_size`` naming the utterances in the batch
    '''
    class EvaluationDataSet(UtteranceContextWindowDataSet):
        '''Append feat_size and utt_id to each sample's tuple'''

        def __getitem__(self, idx):
            feats, alis = super(
                EvaluationDataLoader.EvaluationDataSet, self).__getitem__(idx)
            feat_size = feats.size()[0]
            utt_id = self.utt_ids[idx]
            return feats, alis, feat_size, utt_id

    @staticmethod
    def eval_collate_fn(seq):
        '''Update context_window_seq_to_batch to handle feat_sizes, utt_ids'''
        feats, alis, feat_sizes, utt_ids = zip(*seq)
        feats, alis = context_window_seq_to_batch(zip(feats, alis))
        return feats, alis, tuple(feat_sizes), tuple(utt_ids)

    def __init__(
            self, data_dir, params, file_prefix='', file_suffix='.pt',
            warn_on_missing=True, **kwargs):
        for bad_kwarg in (
                'batch_size', 'sampler', 'batch_sampler', 'shuffle',
                'collate_fn'):
            if bad_kwarg in kwargs:
                raise TypeError(
                    'keyword argument "{}" invalid for {} types'.format(
                        bad_kwarg, type(self)))
        self.data_dir = data_dir
        self.params = params
        self.data_source = self.EvaluationDataSet(
            data_dir, params.context_left, params.context_right,
            file_prefix=file_prefix, file_suffix=file_suffix,
            warn_on_missing=warn_on_missing,
        )
        super(EvaluationDataLoader, self).__init__(
            self.data_source,
            batch_size=params.batch_size,
            shuffle=False,
            collate_fn=self.eval_collate_fn,
            **kwargs
        )
